import pymssql
from ..base import Base
from ..dialect import Dialect

class DialectMssql(Dialect):
	
	@classmethod
	def limit(klass, limit, offset = None):
		"""
		Require SQL Server >= 2012
		"""
		if offset:
			return 'OFFSET {} ROWS FETCH NEXT {} ROWS ONLY'.format(offset, limit), -1
		else:
			return 'TOP {}'.format(limit), 1

class Mssql(Base):

	def __init__(self, *args, **kwargs):
		super(Mssql, self).__init__(*args, **kwargs)
		self.cursor = self.connection.cursor()
		self.dialect(DialectMssql)
	
	def _connect(self, *args, **kwargs):
		arguments = {
			'server' : '.',
			'port'   : 1433
		}
		arguments.update(kwargs)
		return pymssql.connect(**arguments)
		



from .exception import FieldParseError, TableParseError, WhereParseError, UpdateParseError, JoinParseError, LimitParseError, InsertParseError

from . import utils
from .dialect import Dialect
import re

class Term(object):
	"""
	Any part of a sql statement could be a Term
	"""
	def __str__(self):
		raise NotImplementedError('API not implemented: cannot convert to string.')

	def __eq__(self, other):
		return str(self) == str(other)

	def __ne__(self, other):
		return not self == other

	def __hash__(self):
		return hash(str(self))

class Raw(Term):
	"""
	Raw strings
	"""
	def __init__(self, s):
		self.s = str(s)

	def __str__(self):
		return self.s

class Table(Term):
	"""
	Only table or schema.table
	"""
	# a.c(alias)
	REGEX_FROM = r'^\s*((?:[\w_]+\.)?(?:[\w_]+|\*))\s*(?:\(([\w_]+)\))?\s*$'
	# a.c[=] # comment
	REGEX_JOIN = r'^\s*([\w_]+\.)?([\w_]+)\s*(?:\[(.+?)\])?\s*(?:#.*)?$'

	def __init__(self, table, schema = None):
		parts = table.split('.')
		if len(parts) == 1:
			self.table  = table
			self.schema = schema
		elif len(parts) == 2:
			if schema:
				raise TableParseError('Confusing schema specified in arguments "table" and "schema"')
			self.table = parts[1]
			self.schema = parts[0]
		else:
			raise TableParseError('Additional parts in "table"')

	def __str__(self):
		parts = []
		if self.schema:
			parts.append(Builder.DIALECT.quote(self.schema))
		parts.append(Builder.DIALECT.quote(self.table))
		return '.'.join(parts)

	@staticmethod
	def parse(tablestr, context = None):
		if isinstance(tablestr, Term):
			return tablestr
		else:
			if not context:
				return Table(tablestr)
			elif context == 'from':
				m = re.match(Table.REGEX_FROM, tablestr)
				if not m:
					raise TableParseError('Unrecognized table string.')
				table = m.group(1)
				alias = m.group(2)
				return TableFrom(table, alias = alias)
			elif context in ['insert', 'update', 'delete', 'selectinto']:
				return Table(tablestr)
			else:
				raise TableParseError('Unknown table context: {}'.format(context))

class TableFrom(Term):
	def __init__(self, table, alias = None):
		self.table = Table(table)
		self.alias = alias

	def __str__(self):
		return str(self.table) + (' AS ' + Builder.DIALECT.quote(self.alias) if self.alias else '')

class Field(Term):
	"""
	Only field or table.field or schema.table.field
	"""
	# a.b.c(alias)
	REGEX_SELECT  = r'^\s*((?:[\w_]+\.)?(?:[\w_]+\.)?(?:[\w_]+|\*))\s*(?:\|\s*(\.?[\w_]+))?\s*(?:\(([\w_]+)\))?\s*$'
	# a.b.f1, a.b.f2(func1, func2)[=] # comment
	# REGEX_WHERE   = r'^\s*([\w\s_.]+)\s*(?:\(([\w_,]+)\))?\s*(?:\[(.+?)\])?\s*(?:#.*)?$'

	def __init__(self, field = '*', table = None, schema = None):

		parts = field.split('.')
		if len(parts) == 1:
			self.field  = field
			self.table  = table
			self.schema = schema
		elif len(parts) == 2:
			if table:
				raise FieldParseError('Confusing table specified in arguments "field" and "table"')
			self.field = parts[1]
			self.table = parts[0]
			self.schema = schema
		elif len(parts) == 3:
			if table or schema:
				raise FieldParseError('Confusing table/schema specified in arguments "field" and "table"/"schema"')
			self.field = parts[2]
			self.table = parts[1]
			self.schema = parts[0]
		else:
			raise FieldParseError('Additional parts in "field"')

	def __str__(self):
		parts = []
		if self.schema:
			parts.append(Builder.DIALECT.quote(self.schema))
		if self.table:
			parts.append(Builder.DIALECT.quote(self.table))
		parts.append(Builder.DIALECT.quote(self.field))
		return '.'.join(parts)

	def __add__(self, value):
		return Raw(str(self) + '+' + Builder.DIALECT.value(value))

	def __sub__(self, value):
		return Raw(str(self) + '-' + Builder.DIALECT.value(value))

	def __mul__(self, value):
		return Raw(str(self) + '*' + Builder.DIALECT.value(value))

	def __div__(self, value):
		return Raw(str(self) + '/' + Builder.DIALECT.value(value))

	def __mod__(self, value):
		return Raw(str(self) + '%' + Builder.DIALECT.value(value))

	@staticmethod
	def parse(fieldstr, context = None):
		if isinstance(fieldstr, Term):
			return fieldstr
		else:
			if not context:
				return Field(fieldstr)
			elif context == 'select':
				m = re.match(Field.REGEX_SELECT, fieldstr)
				if not m:
					raise FieldParseError('Unrecognized field string: {}'.format(fieldstr))
				field = m.group(1)
				func  = m.group(2)
				alias = m.group(3)
				return FieldSelect(field, alias = alias, func = func)
			elif context == 'group':
				return Field(fieldstr)
			else:
				raise FieldParseError('Unknown field context: {}'.format(context))

class FieldSelect(Term):

	def __init__(self, field = '*', alias = None, func = None):
		"""
		The fields in SELECT substatements
		@params:
			`field`   : The field that applied
			`alias`    : The alias
		"""
		self.field = Field(field)
		self.alias = alias
		self.distinct = func and func.startswith('.')
		self.func  = func[1:] if func and func.startswith('.') else func

	def __str__(self):
		ret = str(self.field)
		if self.func:
			if hasattr(Builder.DIALECT, self.func.lower()):
				ret = getattr(Builder.DIALECT, self.func.lower())(ret, distinct = self.distinct)
			else:
				ret = '{}({}{})'.format(self.func.upper(), 'DISTINCT ' if self.distinct else '', ret)
		if self.alias:
			ret += ' AS ' + Builder.DIALECT.quote(self.alias)
		return ret

class Where(Term):

	def __init__(self, conditions, root = True):
		self.conditions = conditions
		self.root       = root

	def __str__(self):
		sqlitems = []
		for key, val in self.conditions.items():
			# whatever term it is, the value will be ignored
			if isinstance(key, Term):
				sqlitems.append(str(key))
			elif key.split('#')[0].strip().upper() == 'AND':
				if not isinstance(val, (tuple, list, dict)):
					raise WhereParseError('Expect dict or item list/tuple for conditions to be connected by AND: "%s"' % val)
				val = val.items() if isinstance(val, dict) else [
					v if isinstance(v, tuple) else (v, None) for v in val
				]
				whereterms = [str(Where({k:v}, False)) for k, v in val]
				if len(whereterms) == 1:
					sqlitems.append(whereterms[0])
				else:
					sqlitems.append('(%s)' % (' AND '.join(whereterms)))
			elif key.split('#')[0].strip().upper() == 'OR':
				if not isinstance(val, (tuple, list, dict)):
					raise WhereParseError('Expect dict or item list for conditions to be connected by OR: "%s"' % val)
				val = val.items() if isinstance(val, dict) else [
					v if isinstance(v, tuple) else (v, None) for v in val
				]
				whereterms = [str(Where({k:v}, False)) for k, v in val]
				if len(whereterms) == 1:
					sqlitems.append(whereterms[0])
				else:
					sqlitems.append('(%s)' % (' OR '.join(whereterms)))
			else:
				sqlitems.append(str(WhereTerm(key, val)))

		if self.root and len(sqlitems) == 1 and sqlitems[0][0] == '(' and sqlitems[0][-1] == ')':
			# remove the brackets
			return sqlitems[0][1:-1]
		return ' AND '.join(sqlitems)

class WhereTerm(Term):


	REGEX_KEY = r'^\s*(!)?\s*([\w\s_.]+)\s*(?:\|([\w\s_.]+))?\s*(?:\[(.+?)\])?\s*(?:#.*)?$'

	def __init__(self, key, val):
		self.key = key
		self.val = val

	def __str__(self):
		# whatever term it is for the key, ignore the value
		if isinstance(self.key, Term):
			return str(self.key)

		m = re.match(WhereTerm.REGEX_KEY, self.key)
		if not m:
			raise WhereParseError('Unrecognized key in where conditions.')
		ret   = 'NOT ' if m.group(1) else ''
		field = FieldSelect(m.group(2), func = m.group(3))
		oprt  = m.group(4)

		return ret + Builder.DIALECT._operator(oprt, field, self.val)

class Order(Term):

	def __init__(self, orders):
		self.orders = orders

	def __str__(self):
		return ','.join([str(OrderTerm(key, val)) for key, val in self.orders.items()])

class OrderTerm(Term):

	REGEX_KEY  = r'^\s*((?:[\w_]+\.)?(?:[\w_]+\.)?(?:[\w_]+|\*))\s*(?:\|([\w_.]+))?\s*$'
	def __init__(self, key, val):
		m = re.match(OrderTerm.REGEX_KEY, key)
		if not m:
			raise FieldParseError('Unrecognized field in ORDER BY clause.')
		self.field = FieldSelect(m.group(1), func = m.group(2))
		if val is True or val is None:
			val = 'ASC'
		elif val is False:
			val = 'DESC'
		else:
			val = val.upper()
		self.val = val

	def __str__(self):
		return '{} {}'.format(self.field, self.val)

class Limit(Term):

	def __init__(self, limoff):
		if len(limoff) == 1:
			limit  = limoff[0]
			offset = None
		elif len(limoff) == 2:
			limit, offset = limoff
		else:
			raise LimitParseError('LIMIT requires a two integer tuple/list.')

		lim = Builder.DIALECT.limit(limit, offset)
		if isinstance(lim, tuple):
			self.s, self.pos = lim
		else:
			self.s   = lim
			self.pos = -1

	def __str__(self):
		return self.s

class Set(Term):

	def __init__(self, sets):
		self.sets = sets

	def __str__(self):
		return ','.join([str(SetTerm(key, val)) for key, val in self.sets.items()])

class SetTerm(Term):

	REGEX_KEY = r'^\s*((?:[\w_]+\.)?(?:[\w_]+\.)?(?:[\w_]+|\*))\s*(?:\[(.+?)\])?\s*$'
	def __init__(self, key, val):
		m = re.match(SetTerm.REGEX_KEY, key)
		if not m:
			raise UpdateParseError('Unrecognized field in UPDATE SET.')
		self.field = Field(m.group(1))
		self.oprt  = m.group(2)
		self.val   = val

	def __str__(self):
		return Builder.DIALECT._update(self.oprt, self.field, self.val)

class Join(Term):

	def __init__(self, joins, maintable = None):
		self.joins = joins
		self.maintable = maintable

	def __str__(self):
		return ' '.join([str(JoinTerm(key, val, self.maintable)) for key, val in self.joins.items()])

class JoinTerm(Term):

	REGEX_KEY = r'^\s*(?:\[(.+?)\])?\s*((?:[\w_]+\.)?(?:[\w_]+\.)?(?:[\w_]+|\*))\s*(?:\((.+?)\))?\s*$'
	def __init__(self, key, val, maintable = None):
		m = re.match(JoinTerm.REGEX_KEY, key)
		if not m:
			raise JoinParseError('Unrecognized table in JOIN.')
		self.jointype = m.group(1)
		self.table    = TableFrom(m.group(2), alias = m.group(3))
		self.onfields = []
		fieldtable    = self.table.alias or self.table.table
		if isinstance(maintable, TableFrom):
			maintable = maintable.alias or maintable.table
		elif isinstance(maintable, Builder):
			# subquery
			if maintable._subas in [None, True, False]:
				raise JoinParseError('Require alias for subquery to refer to its fields in JOIN ON.')
			else:
				maintable = maintable._subas

		if isinstance(val, (tuple, list)):
			for v in val:
				f1 = Field(v)
				if f1.table:
					raise JoinParseError('Expected table specified in JOIN ON fields.')
				if not maintable:
					raise JoinParseError('Short format of JOIN fields are not allowed without primary table.')
				f2 = Field(v)
				f1.table = fieldtable
				f2.table = maintable
				self.onfields.append((f1, f2))
		elif isinstance(val, dict): # dict
			for k, v in val.items():
				f1 = Field(k)
				f2 = Field(v)

				if f1.table:
					raise JoinParseError('Unexpected table on JOIN ON left field: {}'.format(f1))

				if maintable and f2.table:
					raise JoinParseError('Unexpected table on JOIN ON right field: {}'.format(f2))

				f1.table = fieldtable
				if maintable:
					f2.table = maintable
				self.onfields.append((f1, f2))
		else:
			f1 = Field(val)
			if f1.table:
				raise JoinParseError('Unexpected table specified in JOIN ON fields.')

			if not maintable:
				raise JoinParseError('Short format of JOIN fields are not allowed without primary table.')

			f2 = Field(val)
			f1.table = fieldtable
			f2.table = maintable
			self.onfields.append((f1, f2))

	def __str__(self):
		return '{} {} ON {}'.format(
			Builder.DIALECT._join(self.jointype),
			self.table,
			' AND '.join(['{}={}'.format(key, val) for key, val in self.onfields])
		)

class Builder(Term):

	DIALECT = Dialect

	def __init__(self, dialect = None):
		Builder.DIALECT = dialect or Dialect

		# for join
		self.table  = None
		self.terms  = []
		self._sql   = None
		self._subas = False

	def _select(self, *fields, **kwargs):
		if not fields: fields = ['*']
		distinct = kwargs.get('distinct', False)
		self.terms.append('SELECT DISTINCT' if distinct else 'SELECT')

		fieldterms = [
			Field.parse(field or '*', 'select')
			for field in fields
		]
		self.terms.append(','.join([str(field) for field in fieldterms]))
		return self

	def _sub(self, alias = None):
		"""
		Give a name for subquery
		"""
		self._subas = alias if alias else True
		return self

	def _from(self, *tables):
		self.terms.append('FROM')
		tableterms = [
			Table.parse(table, 'from')
			for table in tables
		]

		self.table = tableterms[0]._subas if isinstance(tableterms[0], Builder) else tableterms[0]
		self.terms.append(','.join([str(table) for table in tableterms]))
		return self

	def _where(self, conditions):
		if not conditions:
			return self
		self.terms.append('WHERE')
		whereterm = Where(conditions)
		self.terms.append(whereterm)
		return self

	def _order(self, orders):
		self.terms.append('ORDER BY')
		self.terms.append(Order(orders))
		return self

	def _limit(self, limoff):
		if not isinstance(limoff, (tuple, list)):
			limoff = [limoff]
		lim = Limit(limoff)
		if lim.pos == -1:
			self.terms.append(lim)
		elif isinstance(lim.pos, int):
			self.terms.insert(lim.pos, str(lim))
		else: # oracle, add where condition
			# check if where is there:
			if 'WHERE' not in self.terms:
				self.terms.append('WHERE')
				self.terms.append(lim)
			else:
				whereindex = self.terms.index('WHERE')
				# TODO: ' AND ' and ' OR ' could also be in subquery or values
				# it's still OK to have brackets, keep it for now
				if ' AND ' in str(self.terms[whereindex + 1]) or ' OR ' in str(self.terms[whereindex + 1]):
					self.terms[whereindex + 1] = '({}) AND ({})'.format(self.terms[whereindex + 1], lim)
				else:
					self.terms.insert(whereindex+2, 'AND ({})'.format(lim))

		return self

	def _union(self, other, all_ = False):
		self.terms.append('UNION')
		if all_:
			self.terms.append('ALL')
		self.terms.append(other)
		return self

	def _exists(self, query):
		self.terms.append('EXISTS')
		self.terms.append(query)
		return self

	def _group(self, fields):
		self.terms.append('GROUP BY')
		if isinstance(fields, Term):
			fields = [fields]
		else:
			fields = utils.alwaysList(fields)
		fieldterms = [
			Field.parse(field, 'group')
			for field in fields
		]
		self.terms.append(','.join([str(field) for field in fieldterms]))
		return self

	def _having(self, conditions):
		if not conditions:
			return self
		self.terms.append('HAVING')
		self.terms.append(Where(conditions))

		return self

	def _insert(self, table, values, fields = None):
		self.terms.append('INSERT INTO')
		self.terms.append(Table(table))
		if not fields:
			if isinstance(values[0], dict):
				fields = values[0].keys()
				for i, value in enumerate(values):
					if i == 0: continue
					if isinstance(value, dict) and set(fields) != set(value.keys()):
						raise InsertParseError('Inconsistent keys in values for INSERT.')
		if fields:
			fieldterms = [Table.parse(field, 'insert') for field in utils.alwaysList(fields)]
			self.terms.append('({})'.format(','.join([str(field) for field in fieldterms])))

		# support INSERT INTO SELECT ...
		if isinstance(values[0], Builder):
			subquery0 = values.pop(0)
			self.terms.append(subquery0)
			for value in values:
				if isinstance(value, Builder):
					if value._subas:
						self.terms.append('UNION ALL')
						value._subas = None
						self.terms.append(value)
					else:
						self.terms.append('UNION')
						self.terms.append(value)
				else:
					raise InsertParseError('All values should be subqueries in INSERT.')
		else:
			insertvals = []
			for value in values:
				if isinstance(value, (tuple, list)):
					insertvals.append(tuple(value))
				elif isinstance(value, dict):
					if fields:
						insertvals.append(tuple(
							value[field] if field in value else None
							for field in fields
						))
					else:
						insertvals.append(tuple(value.values()))
				else:
					raise InsertParseError('Unsupported type for values in INSERT: {}'.format(type(value)))
			self.terms.append('VALUES')
			valterms = []
			for inval in insertvals:
				valterms.append('(%s)' % (','.join([Builder.DIALECT.value(iv) for iv in inval])))
			self.terms.append(','.join(valterms))
		return self

	def _update(self, table):
		self.terms.append('UPDATE')
		self.terms.append(Table.parse(table, 'update'))
		return self

	def _set(self, sets):
		self.terms.append('SET')
		self.terms.append(Set(sets))
		return self

	def _delete(self, table):
		self.terms.append('DELETE FROM')
		self.terms.append(Table.parse(table, 'delete'))
		return self

	def _join(self, joins):
		self.terms.append(Join(joins, self.table))
		return self

	# support select * into newtable from ...
	def _into(self, table):
		self.terms.append('INTO')
		self.terms.append(Table.parse(table, 'selectinto'))
		return self

	def select(self, table, columns = '*', where = None, join = None, distinct = False, newtable = None, sub = None):
		order  = None
		limit  = None
		group  = None
		having = None
		exists = None
		if where and 'ORDER' in where:
			order = where['ORDER']
			del where['ORDER']
		if where and 'LIMIT' in where:
			limit = where['LIMIT']
			del where['LIMIT']
		if where and 'GROUP' in where:
			group = where['GROUP']
			del where['GROUP']
		if where and 'HAVING' in where:
			having = where['HAVING']
			del where['HAVING']
		if where and 'EXISTS' in where:
			exists = where['EXISTS']
			del where['EXISTS']
		if isinstance(columns, Term):
			columns = [columns]
		self._select(*utils.alwaysList(columns), distinct = distinct)
		if newtable:
			self._into(newtable)._from(*utils.alwaysList(table))
		else:
			self._from(*utils.alwaysList(table))
		if join:
			self._join(join)._where(where)
		else:
			self._where(where)
		if order : self._order(order)
		if limit : self._limit(limit)
		if group : self._group(group)
		if having: self._having(having)
		if exists: self._exists(exists)
		if sub   : self._sub(sub)
		return self

	def update(self, table, data, where = None):
		self._update(table)._set(data)._where(where)
		return self

	def delete(self, table, where):
		self._delete(table)._where(where)
		return self

	def insert(self, table, fields, *values):
		"""
		data: list will be treated as fields
		"""
		# table, values, fields
		values2 = []
		if isinstance(fields, dict):
			values2.append(list(fields.values()))
			fields = list(fields.keys())
		elif isinstance(fields, tuple):
			values2.append(tuple(fields))
			fields = None
		elif isinstance(fields, list):
			pass
		else: # assuming fields specified as string
			fields = utils.alwaysList(fields)

		values2.extend([
			value if isinstance(value, tuple) else tuple(value.values())
			for value in values
		])
		self._insert(table, values2, fields)
		return self

	def union(self, *queries):
		queries = list(queries)
		if not self.terms:
			query = queries.pop(0)
			subas = query._subas
			query._subas = None # remove brackets
			self.terms.append(str(query))
			query._subas = subas
		for query in queries:
			if query._subas:
				subas = query._subas
				query._subas = None
				self.terms.append('UNION ALL')
				self.terms.append(str(query))
				query._subas = subas
			else:
				self.terms.append('UNION')
				self.terms.append(query)
		return self

	def sql(self):
		if self._sql:
			return self._sql

		ret = ' '.join(['%s' % t for t in self.terms])
		if self._subas is True:
			ret = '({})'.format(ret)
		elif self._subas:
			ret = '({}) AS {}'.format(ret, Builder.DIALECT.quote(self._subas))

		return ret

	def __str__(self):
		return self.sql()



import sqlite3
from ..base import Base
from ..dialect import Dialect

class DialectSqlite(Dialect):
	pass

class Sqlite(Base):

	def __init__(self, *args, **kwargs):
		database = kwargs.get('database', None)
		if database is not None:
			if database.startswith('file://'):
				database.replace('file://', '')
			kwargs['database'] = database
		super(Sqlite, self).__init__(*args, **kwargs)
		self.cursor = self.connection.cursor()
		self.dialect(DialectSqlite)

	def _connect(self, *args, **kwargs):
		arguments = {
			'database_file'    : ':memory:',
			'timeout'          : 5.0,
			'detect_types'     : 0,
			'isolation_level'  : None,
			'check_same_thread': False,
			#'factory'          : [str][0],
			'cached_statements': 100
		}
		arguments.update(kwargs)
		if 'database' not in arguments:
			arguments['database'] = arguments['database_file']
		del arguments['database_file']

		database = arguments['database']
		if database.startswith('file://'):
			database = database.replace('file://', '')
		arguments['database'] = database

		return sqlite3.connect(**arguments)



from collections import OrderedDict
from .exception import RecordKeyError, RecordAttributeError
from . import utils

class Record(object):
	"""
	A row, from a query, from a database.
	The idea is borrowed from https://github.com/kennethreitz/records
	"""
	def __init__(self, keys, values, readonly = True):
		# sacrifice some efficiency but support setitem, setattr operation
		self.__dict__['_keys']     = keys if readonly else keys[:]
		self.__dict__['_values']   = values if readonly else values[:]
		self.__dict__['_readonly'] = readonly
		# Ensure that lengths match properly.
		assert len(self.keys()) == len(self.values())

	def keys(self):
		"""
		Returns the list of column names from the query.
		"""
		return self.__dict__['_keys']

	def values(self):
		"""
		Returns the list of values from the query.
		"""
		return self.__dict__['_values']

	def __repr__(self):
		return '<Record {}>'.format(self.as_dict())

	def __getitem__(self, key):
		# Support for index-based lookup.
		if isinstance(key, int):
			return self.values()[key]

		# Support for string-based lookup.
		if key in self.keys():
			i = self.keys().index(key)
			if self.keys().count(key) > 1:
				raise RecordKeyError("Record contains multiple '{}' fields.".format(key))
			return self.values()[i]

		raise RecordKeyError("Record contains no '{}' field.".format(key))

	def __setitem__(self, key, val):
		if self.__dict__['_readonly']:
			raise RecordKeyError("Readonly Record does not support setitem operation.")

		if isinstance(key, int):
			self.values()[key] = val
			return
		
		keycount = self.keys().count(key)
		if keycount > 1:
			raise RecordKeyError("Record contains multiple '{}' fields.".format(key))
		elif keycount == 1:
			i = self.keys().index(key)
			self.values()[i] = val
		else: # 0
			self.keys().append(key)
			self.values().append(val)

	def __delitem__(self, key):
		if self.__dict__['_readonly']:
			raise RecordKeyError("Readonly Record does not support setitem operation.")

		# Support for index-based lookup.
		if isinstance(key, int):
			del self.keys()[key]
			del self.values()[key]
			return

		# Support for string-based lookup.
		if key in self.keys():
			i = self.keys().index(key)
			if self.keys().count(key) > 1:
				raise RecordKeyError("Record contains multiple '{}' fields.".format(key))
			del self.keys()[i]
			del self.values()[i]
			return

		raise RecordKeyError("Record contains no '{}' field.".format(key))

	def __getattr__(self, key):
		try:
			return self[key]
		except RecordKeyError as e:
			raise RecordAttributeError(e)

	def __setattr__(self, key, val):
		if self.__dict__['_readonly']:
			raise RecordAttributeError("Readonly Record does not support setitem operation.")

		try:
			self[key] = val
		except RecordKeyError as e:
			raise RecordAttributeError(e)

	def __dir__(self):
		standard = dir(super(Record, self))
		# Merge standard attrs with generated ones (from column names).
		return sorted(standard + [str(k) for k in self.keys()])

	def __eq__(self, other):
		if isinstance(other, OrderedDict):
			return self.as_dict(True) == other
		elif isinstance(other, Record):
			return self.as_dict(True) == other.as_dict(True)
		else:
			return self.as_dict() == dict(other)

	def __ne__(self, other):
		return not self.__eq__(other)

	def __contains__(self, key):
		return key in self.keys()

	def __index__(self, key):
		return self.keys().index(key)

	def get(self, key, default=None):
		"""
		Returns the value for a given key, or default.
		"""
		try:
			return self[key]
		except RecordKeyError:
			return default

	def items(self):
		for i, k in enumerate(self.keys()):
			yield k, self.values()[i]

	def as_dict(self, ordered = False):
		"""
		Returns the row as a dictionary, as ordered.
		"""
		items = zip(self.keys(), self.values())

		return OrderedDict(items) if ordered else dict(items)

	asDict = as_dict


class Records(object):
	"""
	A set of excellent Records from a query.
	"""
	
	def __init__(self, cursor, readonly = True):
		self.meta     = [desc[0] for desc in cursor.description]
		self._cursor  = cursor
		self._allrows = []
		self.pending  = True
		self.readonly = readonly

	def __repr__(self):
		return '<Records: size={}, pending={}>'.format(len(self), self.pending)

	def __iter__(self):
		"""
		Iterate over all rows, consuming the underlying generator
		only when necessary.
		"""
		i = 0
		while True:
			# Other code may have iterated between yields,
			# so always check the cache.
			if i < len(self):
				yield self[i]
			else:
				# Throws StopIteration when done.
				# Prevent StopIteration bubbling from generator, following https://www.python.org/dev/peps/pep-0479/
				try:
					yield next(self)
				except StopIteration:
					return
			i += 1

	def __nonzero__(self):
		return self.first() is not None

	def next(self):
		return self.__next__()

	def __next__(self):
		try:
			nextrow = Record(self.meta, list(next(self._cursor)), readonly = self.readonly)
			self._allrows.append(nextrow)
			return nextrow
		except StopIteration:
			self.pending = False
			raise StopIteration('Records contains no more rows.')

	def __getitem__(self, key):
		is_int = isinstance(key, int)

		# Convert RecordCollection[1] into slice.
		if is_int:
			key = slice(key, key + 1)

		while len(self) < key.stop or key.stop is None:
			try:
				next(self)
			except StopIteration:
				break

		ret = self._allrows[key]
		return ret[0] if is_int else ret

	def __len__(self):
		return len(self._allrows)

	def export(self, format, **kwargs): # pragma: no cover
		"""
		Export the RecordCollection to a given format (courtesy of Tablib).
		"""
		return self.tldata.export(format, **kwargs)

	@property
	def tldata(self):  # pragma: no cover
		"""
		A Tablib Dataset representation of the Records.
		"""
		import tablib
		# Create a new Tablib Dataset.
		data = tablib.Dataset()

		# If the RecordCollection is empty, just return the empty set
		# Check number of rows by typecasting to list
		if len(list(self)) == 0:
			return data

		data.headers = self.meta
		for row in self.all():
			row = utils.reduce_datetimes(row.values())
			data.append(row)

		return data

	def all(self, asdict = False):
		"""
		Returns a list of all rows for the RecordCollection. If they haven't
		been fetched yet, consume the iterator and cache the results.
		@params:
			`asdict`: Whether convert the records to dicts.
				- `False`: don't convert, keep `Record`
				- `True`:  convert to plain dict
				- `ordered`: convert to `OrderedDict`
		"""

		# By calling list it calls the __iter__ method
		if not asdict:
			return list(self)
		elif asdict is True:
			return [r.as_dict() for r in self]
		else:
			return [r.as_dict(True) for r in self]

	def first(self, default = None):
		"""
		Returns a single record for the RecordCollection, or `default`. If
		`default` is an instance or subclass of Exception, then raise it
		instead of returning it.
		"""

		# Try to get a record, or return/raise default.
		try:
			record = self[0]
		except IndexError:
			if isinstance(default, Exception):
				raise default
			return default

		return record





import cx_Oracle
from ..base import Base
from ..dialect import Dialect

class DialectOracle(Dialect):
	pass

class Oracle(Base):

	def __init__(self, *args, **kwargs):
		super(Oracle, self).__init__(*args, **kwargs)
		self.cursor = self.connection.cursor()
		self.dialect(DialectOracle)
	
	def _connect(self, *args, **kwargs):
		arguments = {
			# some default settings
		}
		arguments.update(kwargs)
		return cx_Oracle.connect(**arguments)
		


from setuptools import setup, find_packages

# get version
from os import path
verfile = path.join(path.dirname(__file__), 'medoo', '__init__.py')
with open(verfile) as vf:
    VERSION = vf.readline().split('=')[1].strip()[1:-1]

setup (
	name             = 'medoo',
	version          = VERSION,
	description      = "A lightweight database framework for python",
	url              = "https://github.com/pwwang/pymedoo",
	author           = "pwwang",
	author_email     = "pwwang@pwwang.com",
	license          = "MIT License",
	long_description = 'https://github.com/pwwang/pymedoo',
	packages         = find_packages(),
	install_requires = [
		'six'
    ],
)
